# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Function:
import abc
import time
import uuid
from typing import (
    Any,
)
from typing_extensions import Annotated, Doc
from contextlib import asynccontextmanager

from fastapi import FastAPI, Request
from fastapi.responses import JSONResponse
from fastapi_cdn_host import patch_docs
from starlette.concurrency import iterate_in_threadpool

from zhousflib.web.fastapi.exceptions import EXCEPTION_HANDLERS
# pip install fastapi-cdn-host
"""
from pathlib import Path

from fastapi import FastAPI
from zhousflib.web.fastapi.server import App

from my_log import Loggers


ML_MODELS = {}


class Application(App):

    def __init__(self):
        super().__init__(logger=Loggers, show_docs=False, models=ML_MODELS)

    def init_plugins(self, app: FastAPI, models: dict):
        from fastapi.staticfiles import StaticFiles
        app.mount('/static', StaticFiles(directory=Path(__file__).parent.joinpath("app").joinpath("static")), name='static')


app_sim = Application()

if __name__ == "__main__":
    app_sim.start(app="my_app:app_sim", host="0.0.0.0", port=5006, access_log=True, workers=1)
    
"""


class App(FastAPI, metaclass=abc.ABCMeta):

    def __init__(self, models: dict, show_docs=True, print_request_body=True, print_response_body=True, logger=None, **extra: Annotated[
        Any,
        Doc(
            """
            Extra keyword arguments to be stored in the app, not used by FastAPI
            anywhere.
            """
        ),
    ]):
        extra["lifespan"] = self.lifespan
        # self._exception_handlers = extra["exception_handlers"] if "exception_handlers" in extra else EXCEPTION_HANDLERS
        # extra["exception_handlers"] = self._exception_handlers
        if not show_docs:
            extra["docs_url"] = None
            extra["redoc_url"] = None
            extra["openapi_url"] = None
        super().__init__(**extra)
        self.logger = logger.logger if logger is not None else None
        self.print_request_body = print_request_body
        self.print_response_body = print_response_body
        self.models = models
        if self.logger:
            logger.init_config()

        @self.get("/")
        async def welcome():
            return JSONResponse(content={"message": "hallo!"})

        @self.get("/favicon.ico")
        async def get_favicon():
            return JSONResponse(content={"file": "static/favicon.ico"})

        @self.middleware("http")
        async def interception(request: Request, call_next):
            start_time = time.time()
            request.scope["request_time"] = start_time
            request_id = str(uuid.uuid4())
            if self.logger:
                with self.logger.contextualize(request_id=request_id):
                    self.logger.info("\n")
                    self.logger.info(f"【request url】: {request.url.path}")
                    if self.print_request_body:
                        self.logger.info("【fetch request body...】")
                        body = await request.body()
                        if body:
                            # noinspection PyBroadException
                            try:
                                body_json = await request.json()
                                body = str(body_json)
                                if len(body) > 2048:
                                    body = body[:2048] + "......"
                                self.logger.info(f"【request body】: {body}")
                            except Exception as e:
                                pass
                    self.logger.info("【call next...】")
            response = await call_next(request)
            elapsed_time = f"{time.time() - start_time:.5f}"
            if self.logger:
                with self.logger.contextualize(request_id=request_id):
                    if self.print_response_body:
                        if len(response.headers) > 0:
                            content_type = response.headers.get("content-type", "")
                            content_disposition = response.headers.get("content-disposition", None)
                            if content_type == "application/json" and content_disposition is None:
                                response_body = [chunk async for chunk in response.body_iterator]
                                response.body_iterator = iterate_in_threadpool(iter(response_body))
                                if len(response_body) > 0:
                                    body = str(response_body[0].decode())
                                    if len(body) > 2048:
                                        body = body[:2048] + "......"
                                    self.logger.info(f"【response body】: {body}")
                    self.logger.info(f"【X-Process-Time】: {elapsed_time}s")
            response.headers["X-Process-Time"] = elapsed_time
            response.headers["X-Request-ID"] = request_id
            return response

    @asynccontextmanager
    async def lifespan(self, app: FastAPI):
        if self.logger:
            self.logger.info("init plugins.")
        patch_docs(app)
        self.init_plugins(app, self.models)
        yield
        self.models.clear()

    @abc.abstractmethod
    def init_plugins(self, app: FastAPI, models: dict):
        pass

    @staticmethod
    def start(app: str, host="0.0.0.0", port=5006, workers=1, **kw):
        import uvicorn
        kw["app"] = app
        kw["host"] = host
        kw["port"] = port
        kw["workers"] = workers
        config = uvicorn.Config(**kw)
        server = uvicorn.Server(config)
        server.run()





# -*- coding: utf-8 -*-
# @Author  : zhousf-a
# @Function:
"""
uninstall numpy, pandas, matplotlib, seaborn
pip install matplotlib==3.7.3
pip install seaborn==0.11.0

[question]
Backend TkAgg is interactive backend. Turning interactive mode on.
[solution]
import matplotlib
matplotlib.use('TkAgg')
"""
from pathlib import Path
import pandas as pd
import numpy as np
import seaborn as sns
from pandas import DataFrame
import matplotlib.pyplot as plt
from matplotlib.font_manager import FontProperties
from zhousflib.font import Font_SimSun
from zhousflib.util import color_util


def draw_histogram(df: DataFrame, title: str, font_size=12, x_label: str = None, y_label: str = "Count", colors: list = None):
    """
    绘制直方图
    :param df: 数据
    :param title: 显示标题
    :param font_size: 显示字体大小
    :param x_label: x轴名换
    :param y_label: y轴名称
    :param colors: ["red", "blue"]
    :return:
    """
    df = df.fillna(np.NaN)
    font = FontProperties()
    font.set_file(Font_SimSun)
    font.set_size(font_size)
    plt.title(title, fontsize='large', fontproperties=font)
    columns = df.columns
    sns.set(font='SimHei', font_scale=0.8, style="darkgrid")
    if colors:
        assert len(colors) >= len(columns), f"colors length is {len(colors)} but columns length is {len(columns)}"
    else:
        colors = color_util.generate_hex_colors(len(columns))
    for i, column in enumerate(columns):
        sns.histplot(data=df, x=column, color=colors[i], label=f"{column}:{len(df[column].dropna())}", kde=True)
    plt.xlabel(x_label, fontproperties=font)
    plt.ylabel(y_label,  fontproperties=font)
    plt.legend()
    plt.show()


if __name__ == "__main__":
    # excel_file = Path(r'D:\workspace\ZhousfLib\zhousflib\data.xlsx')
    # df1_ = pd.read_excel(str(excel_file), header=0)
    df2_ = DataFrame({
            "相似": pd.Series([0.98, 0.82, 0.92, 1.0, 0.88, 0.9, 0.96, 0.95, 0.76, 0.86, 0.86, 0.98, 0.92, 0.99, 0.94, 0.84, 0.9, 0.98, 0.91, 0.97, 0.88, 0.9, 0.93, 1.0, 1.0, 1.0, 0.98, 0.98, 0.91, 0.9, 0.95, 0.71, 0.85, 0.97, 0.84, 0.85, 0.85, 0.87, 0.87, 0.88, 0.88, 0.89, 0.9, 0.95, 0.95, 0.96, 0.97, 1.0, 0.76, 0.77, 0.63, 0.58, 0.68, 0.83, 0.85, 0.81, 0.88, 0.86, 0.95, 0.9, 0.9, 0.94, 0.96, 0.98, 0.83, 0.95, 0.92, 0.95, 0.9, 0.91]),
            "不相似": pd.Series([0.33, 0.13, 0.18, 0.17, 0.0, 0.35, 0.18, 0.4, 0.0, 0.16, 0.4, 0.33, 0.23, 0.2, 0.11, 0.18, 0.29, 0.3, 0.14, 0.14, 0.27, 0.19, 0.13, 0.33, 0.27, 0.22, 0.0, 0.37, 0.22, 0.32, 0.2, 0.25, 0.14, 0.41, 0.11, 0.43, 0.16, 0.36, 0.17, 0.22, 0.23, 0.2, 0.22, 0.34, 0.13, 0.42, 0.35, 0.31, 0.26, 0.31, 0.35, 0.24, 0.29, 0.22, 0.0, 0.0, 0.12, 0.24, 0.26, 0.18, 0.1, 0.39, 0.37, 0.86, 0.87, 0.88, 0.45, 0.27, 0.5, 0.0])
    })
    draw_histogram(df=df2_, title="直方图", x_label="置信度", y_label="数量")
    pass



# -*- coding:utf-8 -*-
# Author:      zhousf
# File:        interceptor_util.py
# Description:  拦截器
import functools


def intercept(before=None, after=None):
    """
    拦截器
    :param before: 过滤器
    :param after: 过滤器
    def before(*args):
        if args[0] == "0":
            return True, '拦截'
        return False, '不拦截'
    :return:
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*chain, **kw):
            if before is not None:
                result_before = before(chain)
                if isinstance(result_before, tuple):
                    need_intercept, msg = result_before
                else:
                    need_intercept = result_before
                    msg = "interceptor"
                if need_intercept:
                    return msg
                result = func(msg, **kw)
            else:
                result = func(chain, **kw)
            return result if after is None else after(result)
        return wrapper

    return decorator


def aop(before=None, after=None):
    def run(run_func, *args, **kw):
        func_ = args[0]
        if len(args) > 1:
            params = args[1:]
            if run_func is not None:
                if len(kw) > 0:
                    run_func(func_, *params, **kw)
                else:
                    run_func(func_, *params)
        else:
            if run_func is not None:
                if len(kw) > 0:
                    run_func(func_, **kw)
                else:
                    run_func(func_)

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kw):
            run(before, *args, **kw)
            run(func, *args, **kw)
            run(after, *args, **kw)
        return wrapper

    return decorator


"""
if __name__ == "__main__":
    class Demo:

        def __init__(self):
            pass
        def before(self, msg):
            print(f"before {msg}")
            msg[0] += "_before"
            msg[1] += 1

        def after(self, msg):
            print(f"after {msg}")
            msg[0] += "_after"
            msg[1] += 1

        @aop(before=before, after=after)
        def test(self, msg):
            print(f"test {msg}")
            msg[0] += "_test"
            msg[1] += 1


    parmas = ["hello", 1]
    Demo().test(parmas)
    print(parmas)
"""





# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Function:

import os


class DockerBuild:

    def __init__(self, docker_hub_domain: str = None, docker_hub_namespace: str = None):
        self.docker_hub_domain = docker_hub_domain
        self.docker_hub_namespace = docker_hub_namespace
        self.docker_hub_url = f"{docker_hub_domain}/{docker_hub_namespace}" if docker_hub_domain and docker_hub_namespace else None
        if self.docker_hub_url:
            print("正在登录docker仓库")
            os.system(f"docker login {self.docker_hub_domain}")
        pass

    @staticmethod
    def __command(command: str, command_tip: str):
        print(f"正在{command_tip}")
        ret = os.system(command)
        assert ret == 0, f"{command_tip}失败: {ret}"
        print(f"{command_tip}成功")

    def build_image(self, dockerfile: str, name: str, version: str, command_tip: str, push=False):
        image_name = f"{name}:{version}"
        self.__command(command=f"docker build --network=host -f {dockerfile} -t {image_name} .", command_tip=command_tip)
        if self.docker_hub_url and push:
            self.push_image(image_name)
        return self

    def pull_image(self, tag: str, command_tip: str):
        self.__command(command=f"docker pull {self.docker_hub_url}/{tag}", command_tip=command_tip)
        return self

    def push_image(self, current_image_name: str, tag_image_name: str = None):
        """
        推送镜像
        :param current_image_name: 当前镜像名称（包括版本号）
        :param tag_image_name: 给镜像打标签（包括版本号），即重命名，默认空
        :return:
        """
        if self.docker_hub_url:
            if tag_image_name is None:
                self.__command(command=f"docker tag {current_image_name} {self.docker_hub_url}/{current_image_name}", command_tip="镜像打标签")
                self.__command(command=f"docker push {self.docker_hub_url}/{current_image_name}", command_tip="推送镜像")
            else:
                self.__command(command=f"docker tag {current_image_name} {self.docker_hub_url}/{tag_image_name}", command_tip="镜像打标签")
                self.__command(command=f"docker push {self.docker_hub_url}/{tag_image_name}", command_tip="推送镜像")
        return self


if __name__ == "__main__":
    _docker_hub_domain = "packages.xxx.com"
    _docker_hub_namespace = "docker-xxx"
    # _docker_hub_domain = "registry.cn-beijing.aliyuncs.com"
    # _docker_hub_namespace = "zhousf-ai"
    build = (DockerBuild(docker_hub_domain=_docker_hub_domain, docker_hub_namespace=_docker_hub_namespace)
             # .pull_image(tag="nvidia/cuda:11.3.1-cudnn8-devel-ubuntu20.04", command_tip="拉取基础镜像")
             # .pull_image(tag="registry.baidubce.com/paddlepaddle/paddle:2.3.1-gpu-cuda11.2-cudnn8 ", command_tip="拉取基础镜像")
             # .pull_image(tag="nvidia/cuda:11.6.1-cudnn8-runtime-ubuntu20.04", command_tip="拉取基础镜像")
             # .build_image(dockerfile="Dockerfile_env_cpu", name="zhousf_fd_py39_cpu", version="1.0.7", push=False, command_tip="构建环境镜像")
             # .build_image(dockerfile="Dockerfile_env_gpu", name="zhousf_fd_py39_cuda11.6.1-cudnn8", version="1.0.7", push=False, command_tip="构建环境镜像")
             .push_image(current_image_name="pdf_omni:web_py3.10_cuda11.8_cudnn8.6_v0.3.1")
             # .push_image(current_image_name="pdf_omni:env_py3.10_cuda11.8_cudnn8.6_v0.3")
             # .push_image(current_image_name="zhousf_fd_py39_cuda11.6.1-cudnn8:1.0.7", tag_image_name="fast_infer:py3.9_cuda11.6.1_cudnn8_v1.0.7")
             # .build_image(dockerfile="Dockerfile_pro", name="zhousf_image_classify_gpu", version="0.1", push=True, command_tip="构建工程镜像")
             )



# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Date    : 2023/11/1 
# @Function:
import os
import base64
import shutil
import imghdr
import hashlib
import traceback
from pathlib import Path
from PIL import Image


def overwrite_folder(from_dir: Path, to_dir: Path):
    """
    覆盖目录，覆盖已有文件
    :param from_dir: 目录
    :param to_dir: 被覆盖目录
    :return:
    """
    for file in from_dir.rglob("*.*"):
        shutil.copy(file, to_dir.joinpath(file.parent.name))


def md5(file_path: Path):
    """
    文件转md5
    :param file_path: 文件路径
    """
    with file_path.open('rb') as f:
        return hashlib.md5(f.read()).hexdigest()


def get_base64(file_path: Path, contain_file_name=False, split_char=","):
    """
    文件转base64
    :param file_path: 文件路径
    :param contain_file_name: 是否包含文件名称
    :param split_char: 分隔符
    :return: 'a.jpg,iVBORw0KGgoAAAANSUhEUgAABNcAAANtCAYAAACzHZ25AAA.....'
    """
    with file_path.open('rb') as infile:
        s = infile.read()
    base64_str = base64.b64encode(s).decode("utf-8")
    if contain_file_name:
        base64_str = file_path.name + split_char + base64_str
    return base64_str


def rename_images_with_md5(file_dir: Path):
    """
    图片MD5重命名
    :param file_dir:
    :return:
    """
    images = [file for file in file_dir.rglob("*.*") if imghdr.what(file)]
    for image in images:
        md5_str = md5(image)
        img_arr = Image.open(image)
        suffix = image.suffix
        if hasattr(img_arr, "format") and img_arr.format.lower() in ["jpg", "png"]:
            suffix = "." + img_arr.format
        suffix = suffix.lower()
        if suffix not in [".jpg", ".png"]:
            suffix = ".png"
        img_arr.close()
        new_name = image.parent.joinpath("{0}{1}".format(md5_str, suffix))
        if image.name == new_name.name:
            continue
        if new_name.exists():
            image.unlink()
            continue
        os.rename(image, new_name)
    return len(images)


def drop_error_image(file_dir: Path, min_size=20):
    """
    删除无效图片
    :param file_dir:
    :param min_size: 图片最小尺寸，小于则删除
    :return:
    """
    images = [file for file in file_dir.rglob("*.*") if imghdr.what(file)]
    for image in images:
        img_arr = Image.open(image)
        try:
            if img_arr is None:
                print(image)
                image.unlink()
                continue
            if min(img_arr.size) < min_size:
                img_arr.close()
                print(image)
                image.unlink()
                continue
        except Exception as e:
            print(e)



# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Function:  pip install python-dotenv
import time
import shutil
import traceback
import multiprocessing
import datetime as dt
from datetime import datetime
from loguru import logger
from dotenv import dotenv_values
from configure import LogDir


config_env = dotenv_values(".env.dev")

gunicorn_dir = LogDir.joinpath("gunicorn")
if not gunicorn_dir.exists():
    gunicorn_dir.mkdir(parents=True, exist_ok=True)


bind = "{0}:{1}".format(config_env.get("HOST"), config_env.get("PORT"))
# 进程文件
pidfile = '{0}/gunicorn.pid'.format(gunicorn_dir)
# 进程数
workers = config_env.get("WORKERS")
# 进程名称
proc_name = "zhousf_project"
# 每个进程开启的线程数
threads = multiprocessing.cpu_count() * 2
# 设置后台守护进程
daemon = config_env.get("DAEMON")
# 设置最大并发量
worker_connections = 50
# 超时
timeout = 1800
# 工作模式协程
worker_class = 'uvicorn.workers.UvicornWorker'
# 日志级别
loglevel = 'warning'
# 访问日志格式
access_log_format = '%(t)s %(p)s %(h)s "%(r)s" %(s)s %(L)s %(b)s %(f)s" "%(a)s"'
# 访问日志文件
accesslog = "{0}/gunicorn_access.log".format(gunicorn_dir)
# 错误日志文件
errorlog = "{0}/gunicorn_error.log".format(gunicorn_dir)


def delete_log_file(keep_day: int = 7):
    logger.success("Start the scheduled task to delete log files")
    log_file = LogDir.joinpath("delete_log_file.txt")
    msg = [datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')]
    delete_count = 0
    # noinspection PyBroadException
    try:
        business_dir = LogDir.joinpath("business")
        if not business_dir.exists():
            msg.append(f"Dir not found: {business_dir}")
            return
        business_dirs = [d for d in LogDir.joinpath("business").iterdir()]
        if len(business_dirs) == 0:
            return
        current_date = datetime.now().date()
        available_date = []
        for date_ in [current_date - dt.timedelta(days=i) for i in range(keep_day)]:
            available_date.append(str(date_).replace("-", "_"))
        for business_dir in business_dirs:
            date_dirs = [d for d in business_dir.iterdir()]
            if len(date_dirs) == 0:
                continue
            for date_dir in date_dirs:
                if date_dir.name in available_date:
                    continue
                msg.append(str(date_dir))
                logger.info("delete log dir: {0}".format(str(date_dir)))
                delete_count += 1
                shutil.rmtree(date_dir, ignore_errors=True)
    except Exception as e:
        msg.append(str(traceback.print_exc()))
    finally:
        logger.info(f"Delete {delete_count} directory in total")
        msg.append(f"Delete {delete_count} directory in total")
        msg.append("\n")
        with log_file.open("a+", encoding="utf-8") as f:
            f.write("\n".join(msg))


def delete_task(keep_day: int = 30):
    from db import db_task
    logger.success("Start the scheduled task to delete tasks")
    log_file = LogDir.joinpath("delete_task.txt")
    msg = [datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')]
    try:
        tasks = db_task.query_all()
        delete_count = 0
        task_ids = []
        if len(tasks) > 0:
            for task_id in tasks.keys():
                task = tasks.get(task_id)
                create_time = task.get("create_time")
                date1 = datetime.fromtimestamp(create_time)
                date2 = datetime.fromtimestamp(time.time())
                delta = (date2 - date1).days
                if delta > keep_day:
                    db_task.delete(task.get("task_id"))
                    delete_count += 1
                    task_ids.append(task.get("task_id"))
        msg.append(f"Delete {delete_count} task in total: {task_ids}")
        logger.info(f"Delete {delete_count} task in total: {task_ids}")
        msg.append("\n")
    except Exception as e:
        msg.append(f"Error: {e}")
    with log_file.open("a+", encoding="utf-8") as f:
        f.write("\n".join(msg))


def when_ready(server):
    logger.success("Gunicorn on ready.")
    from apscheduler.schedulers.background import BackgroundScheduler
    scheduler = BackgroundScheduler()
    scheduler.add_job(func=delete_log_file, trigger='cron', id="delete_log", max_instances=1, hour=1, minute=59, misfire_grace_time=900)
    scheduler.add_job(func=delete_task, trigger='cron', id="delete_task", max_instances=1, hour=1, minute=30, misfire_grace_time=900)
    scheduler.start()
    logger.success("Scheduler on ready.")


def on_exit(server):
    pass




# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Function:
from starlette.exceptions import HTTPException
from fastapi.responses import JSONResponse
from fastapi.exceptions import RequestValidationError


async def http_exception_handler(request, exc):
    return JSONResponse(dict(resultCode=exc.status_code, errorString=exc.detail, status="Failed"))


async def validate_exception_handler(request, exc):
    return JSONResponse(dict(resultCode=400, errorString=exc.errors(), status="Failed"))


async def global_exception_handler(request, exc):
    return JSONResponse(dict(resultCode=400, errorString=str(exc), status="Failed"))


EXCEPTION_HANDLERS = {
    HTTPException: http_exception_handler,
    RequestValidationError: validate_exception_handler,
    Exception: global_exception_handler
}



# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Function:
"""
【question】
Dockerfile中wget总是报443
【solution】
在Dockerfile中增加：
RUN apt-get update && apt-get install -y ca-certificates wget
RUN update-ca-certificates

【question】
Docker中的bash文件运行报错：bash \r；该错误由于在windows运行环境下自动追加\r结尾符导致的
【solution】
在Dockerfile中增加：
RUN apt-get update && \
    apt-get install -y --no-install-recommends \
        dos2unix \
RUN dos2unix /[your_bash].sh
"""


# -*- coding:utf-8 -*-
# Author:  zhousf
# Description:
import cv2
import numpy as np
from pathlib import Path

from zhousflib.decorator import interceptor_util


def get_binary(image_path: Path, **kwargs):
    """
    获取二值化
    :param image_path:
    :param kwargs: 二值化参数
    :return:
    """
    assert image_path.exists(), "image file is not exists: {0}".format(image_path)
    img = cv2.imread(str(image_path), cv2.IMREAD_GRAYSCALE)
    assert img is not None, "image is None"
    if len(kwargs) > 0:
        binary = cv2.adaptiveThreshold(src=img, **kwargs)
    else:
        binary = cv2.adaptiveThreshold(src=img,
                                       maxValue=255,  # 大于阈值后设定的值
                                       adaptiveMethod=cv2.ADAPTIVE_THRESH_MEAN_C,  # 自适应方法
                                       thresholdType=cv2.THRESH_BINARY,  # 同全局阈值法中的参数一样
                                       blockSize=11,  # 方阵（区域）大小
                                       C=15)  # 常数项
    return binary


def rotate(image, angle, show=False):
    """
    顺时针旋转图片
    :param image: 图片
    :param angle: 旋转角度
    :param show: 是否显示
    :return: np.array
    """
    if isinstance(image, str):
        image = cv2.imread(image)
    if isinstance(image, np.ndarray):
        # grab the dimensions of the image and then determine the
        # center
        (h, w) = image.shape[:2]
        (cX, cY) = (w // 2, h // 2)

        # grab the rotation matrix (applying the negative of the
        # angle to rotate clockwise), then grab the sine and cosine
        # (i.e., the rotation components of the matrix)
        M = cv2.getRotationMatrix2D((cX, cY), -angle, 1.0)
        cos = np.abs(M[0, 0])
        sin = np.abs(M[0, 1])

        # compute the new bounding dimensions of the image
        nW = int((h * sin) + (w * cos))
        nH = int((h * cos) + (w * sin))

        # adjust the rotation matrix to take into account translation
        M[0, 2] += (nW / 2) - cX
        M[1, 2] += (nH / 2) - cY

        # perform the actual rotation and return the image
        rotated = cv2.warpAffine(image, M, (nW, nH))
        if show:
            cv2.imshow("{0} rotated".format(angle), rotated)
            cv2.waitKey(0)
        return rotated
    return None


def _read(chain):
    image, show = chain
    if isinstance(image, str):
        image = cv2.imread(image)
    if image is None:
        return True, image
    return False, (image, show)


def _show(chain):
    title, image, show = chain
    if show:
        cv2.imshow(title, image)
        cv2.waitKey(0)
        cv2.destroyAllWindows()
    return image


"""
以下函数为示例函数
"""


@interceptor_util.intercept(before=_read, after=_show)
def img_binary(chain):
    """ 二值化 """
    image, show = chain
    if len(image.shape) == 3:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    block_size = 11  # 分割计算的区域大小，取奇数
    binary_ = cv2.adaptiveThreshold(image, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY_INV, block_size, 15)
    return "binary", binary_, show


@interceptor_util.intercept(before=_read, after=_show)
def img_bilateral_filter(chain):
    """ 双边滤波 """
    image, show = chain
    bilateral_filter_ = cv2.bilateralFilter(image, 7, 75, 75)
    return "bilateral_filter", bilateral_filter_, show


@interceptor_util.intercept(before=_read, after=_show)
def img_erosion(chain):
    """ 腐蚀 """
    image, show = chain
    kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (20, 20))
    erosion_ = cv2.erode(image, kernel)
    return "erosion", erosion_, show


@interceptor_util.intercept(before=_read, after=_show)
def img_dilate(chain):
    """ 膨胀 """
    image, show = chain
    kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (20, 20))
    dilate_ = cv2.dilate(image, kernel)
    return "dilate", dilate_, show


@interceptor_util.intercept(before=_read, after=_show)
def img_clahe(chain):
    """ 直方图均衡化 """
    image, show = chain
    clahe = cv2.createCLAHE(clipLimit=3.0, tileGridSize=(5, 5))
    lab = cv2.cvtColor(image, cv2.COLOR_BGR2LAB)
    l, a, b = cv2.split(lab)
    l2 = clahe.apply(l)
    lab = cv2.merge((l2, a, b))
    clahe_ = cv2.cvtColor(lab, cv2.COLOR_LAB2BGR)
    return "clahe", clahe_, show


@interceptor_util.intercept(before=_read, after=_show)
def img_remove_background(chain):
    """ 去除背景 """
    image, show = chain
    if len(image.shape) == 3:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    binary = cv2.adaptiveThreshold(image, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY_INV, 25, 15)
    se = cv2.getStructuringElement(cv2.MORPH_RECT, (1, 1))
    se = cv2.morphologyEx(se, cv2.MORPH_CLOSE, (2, 2))
    mask = cv2.dilate(binary, se)
    mask1 = cv2.bitwise_not(mask)
    binary = cv2.bitwise_and(image, mask)
    result = cv2.add(binary, mask1)
    return "remove_background", result, show


if __name__ == "__main__":
    image_file = r"C:\Users\zhousf-a\Desktop\0\07398c4c-efb238518a73537efa4b212fceaf7a6f_76_ero.jpg"
    image_file = cv2.imread(image_file, 0)
    # binary = img_binary(image_file, False)
    # dilate = img_dilate(binary, True)
    # erosion = img_erosion(dilate, True)
    # cv2.imwrite(r"C:\Users\zhousf-a\Desktop\0\3_vis_erosion.jpg", erosion)





# -*- coding:utf-8 -*-
# Author:  zhousf
# Description: 相似度计算
# pip install imagehash
# pip install scikit-image
# pip install scikit-learn
import time
from pathlib import Path

import cv2
import imagehash
from PIL import Image
from sklearn import metrics
from skimage.metrics import structural_similarity
from skimage.metrics import mean_squared_error

from zhousflib.image import read

"""
【相似度算法】
（1）传统算法
· 均方误差(Mean Squared Error，MSE)：计算两张图片像素之间的差异，并将这些差异的平方求和，最后除以像素数量得到MSE。MSE值越小，图片越相似。MSE算法只考虑像素级别的差异，可能无法准确地捕捉图像的纹理、结构等细节。
· 结构相似性指数（Structural Similarity Index，SSIM）：通过比较两张图片的亮度、对比度和结构等方面的相似性，给图片打分。SSIM值在0到1之间，1表示两张图片完全相同。SSIM可以更好地检测出字形上的细微差异。
· 峰值信噪比（Peak Signal-to-Noise Ratio，PSNR）：计算两张图片像素之间的差异，并将这些差异的平方求和后再取对数。PSNR值越高，图片越相似。PSNR常常被用来评估图像压缩、图像处理、图像修复等技术的质量。
· 直方图比较（Histogram Comparison）：将两张图片的颜色直方图进行比较，以确定它们的相似程度。直方图只考虑了颜色分布而忽略了纹理和结构等因素。
· 感知哈希（Perceptual Hash，PHash）：通过对图片进行哈希算法，将图片转换为二进制字符串，然后比较字符串之间的汉明距离（Hamming Distance），以确定图片的相似度。哈希算法主要保留了低频分量，存在一定的局限性。
· 互信息：互信息是一种用于衡量两个随机变量之间相互依赖关系的指标，可以用于计算图片的相似度。互信息算法可能对图像的纹理、结构等细节不敏感。
· 特征匹配：使用SIFT/SURF/ORB等特征描述子代表图像特征，再使用欧氏距离/汉明距离等计算相似度指标来衡量两张图片的相似度，最后使用RANSAC等筛选算法提高匹配的准确性。
tips:
当应用场景的图片存在复印件或颜色变换时则不建议使用直方图算法
当应用场景的图片可能存在人为噪声时则不建议使用均方误差和峰值信噪比算法
效果参考：
互信息 -> 结构相似性指数 -> 均方误差 -> 峰值信噪比 -> 感知哈希 -> 直方图
（2）深度学习算法
· 孪生网络：例如人脸识别
· SimGNN：基于图神经网络（Graph Neural Network，GNN）的一种模型，用于处理图数据的相似度计算任务。常用于推荐系统中的物品相似度计算、文本匹配中的句子相似度计算等。
· Graph kernel：Graph kernels（图核）是一类用于计算图数据相似度的方法

【相似度度量算法】
· 余弦相似度：余弦相似度是一种常用的相似度计算方法，特别适用于文本数据的相似度计算。它通过计算两个向量的夹角余弦值来衡量它们之间的相似程度。夹角余弦值越接近1，表示两个向量越相似；夹角余弦值越接近0，表示两个向量越不相似。
· 欧氏距离：欧氏距离是一种用来衡量两个向量之间的距离的方法。它计算的是两个向量之间的直线距离，即两个向量之间的欧氏距离越小，表示它们之间的相似程度越高。
· 汉明距离：汉明距离是一种用来衡量两个等长字符串之间的差异的方法。它计算的是两个字符串之间对应位置不同的字符个数，即汉明距离越小，表示两个字符串越相似。
· 编辑距离：编辑距离是一种用来衡量两个字符串之间的差异的方法。它计算的是通过插入、删除和替换操作将一个字符串转换为另一个字符串所需要的最小操作次数，即编辑距离越小，表示两个字符串越相似。
· Jaccard杰卡德系数：Jaccard相似系数是一种用来衡量两个集合之间的相似性的方法。它计算的是两个集合交集的大小与并集的大小之间的比值，即Jaccard相似系数越大，表示两个集合越相似。
· Pearson皮尔逊相关系数：皮尔逊相关系数是一种用来衡量两个变量之间相关程度的方法。它计算的是两个变量之间的协方差与它们各自标准差的乘积之间的比值，即皮尔逊相关系数越接近1或-1，表示两个变量之间越相关。
"""


def ___read(image1: Path, image2: Path, resize: tuple = None, check_dimensions=True, gray=False):
    """
    :param image1:
    :param image2:
    :param resize:
    :param check_dimensions:
    :param gray:
    :return:
    """
    image1 = read(image1)
    image2 = read(image2)
    # 尺寸标准化
    if resize:
        image1 = cv2.resize(image1, resize)
        image2 = cv2.resize(image2, resize)
    # 灰度化
    if gray:
        image1 = cv2.cvtColor(image1, cv2.COLOR_BGR2GRAY)
        image2 = cv2.cvtColor(image2, cv2.COLOR_BGR2GRAY)
    if check_dimensions:
        assert image1.shape == image2.shape, "Input images must have the same dimensions."
    return image1, image2


def compare_mse(image1: Path, image2: Path, resize: tuple = None):
    """
    计算均方误差比较相似度
    :param image1:
    :param image2:
    :param resize: 尺寸标准化
    :return:
    """
    image1, image2 = ___read(image1, image2, resize=resize, check_dimensions=True, gray=True)
    return mean_squared_error(image1, image2)


def compare_psnr(image1: Path, image2: Path, resize: tuple = None):
    """
    计算峰值信噪比
    :param image1:
    :param image2:
    :param resize: 尺寸标准化
    :return:
    """
    image1, image2 = ___read(image1, image2, resize=resize, check_dimensions=True, gray=True)
    # 亮度统一
    image1 = cv2.normalize(image1, None, 0, 255, cv2.NORM_MINMAX)
    image2 = cv2.normalize(image2, None, 0, 255, cv2.NORM_MINMAX)
    # from skimage.metrics import peak_signal_noise_ratio
    # return peak_signal_noise_ratio(image1, image2)  # 稳健性不如cv2
    return cv2.PSNR(image1, image2)


def compare_ssim(image1: Path, image2: Path, resize: tuple = None, show_diff=False):
    """
    计算结构相似性指数
    :param image1:
    :param image2:
    :param resize: 尺寸标准化
    :param show_diff: 显示结构不相似的区域
    :return:
    """
    image1, image2 = ___read(image1, image2, resize=resize, check_dimensions=True, gray=True)
    score, diff = structural_similarity(image1, image2, full=True)
    if show_diff:
        cv2.imshow("ssim", diff)
        cv2.waitKey(0)
        cv2.destroyAllWindows()
    return score, diff


def compare_hist(image1: Path, image2: Path, resize: tuple = None):
    """
    计算直方图相似性
    :param image1:
    :param image2:
    :param resize: 尺寸标准化
    :return:
    """
    image1, image2 = ___read(image1, image2, resize=resize, check_dimensions=True, gray=True)
    hist1 = cv2.calcHist([image1], [0], None, [256], [0, 256])
    hist2 = cv2.calcHist([image2], [0], None, [256], [0, 256])
    return cv2.compareHist(hist1, hist2, cv2.HISTCMP_CORREL)


def compare_dhash(image1: Path, image2: Path, hash_size=8, resize: tuple[int, int] = None):
    """
    感知hash
    :param image1:
    :param image2:
    :param hash_size:
    :param resize: 尺寸标准化
    :return:
    """
    image1 = Image.open(image1)
    image2 = Image.open(image2)
    if resize:
        image1 = image1.resize(resize, resample=Image.BILINEAR)
        image2 = image2.resize(resize, resample=Image.BILINEAR)
    image1 = imagehash.average_hash(image1, hash_size=hash_size)
    image2 = imagehash.average_hash(image2, hash_size=hash_size)
    return hamming_distance(image1, image2, hash_size=hash_size)


def hamming_distance(hash1, hash2, hash_size=8):
    """
    汉明距离
    """
    distance = hash1 - hash2
    similarity = 1 - (distance / (hash_size * hash_size))
    return similarity


def compare_mutual_info(image1: Path, image2: Path, resize: tuple = None):
    """
    计算互信息相似度
    :param image1:
    :param image2:
    :param resize: 尺寸标准化
    :return:
    """
    image1, image2 = ___read(image1, image2, resize=resize, check_dimensions=True, gray=False)
    return metrics.normalized_mutual_info_score(image1.reshape(-1), image2.reshape(-1))


if __name__ == "__main__":
    # 加载测试图片
    img1 = Path(r'C:\Users\zhousf-a\Desktop\data\1_difficult_sample\29_0.82\0b01f464653b924947aac1e1ba9b98f0.jpg')
    img2 = Path(r'C:\Users\zhousf-a\Desktop\data\1_difficult_sample\29_0.82\014b12e54e01cdd23a2c205a73d25e40.jpg')

    start_time = time.time()
    score_ = compare_mutual_info(img1, img2, resize=(200, 200))
    print('互信息的相似度：', score_, '耗时：', time.time() - start_time)

    start_time = time.time()
    score_, _ = compare_ssim(img1, img2, resize=(200, 200))
    print('SSIM的相似度：', score_, '耗时：', time.time() - start_time)

    start_time = time.time()
    score_ = compare_mse(img1, img2, resize=(200, 200))
    print('MSE的相似度：', score_, '耗时：', time.time() - start_time)

    start_time = time.time()
    score_ = compare_psnr(img1, img2, resize=(200, 200))
    print('PSNR的相似度：', score_, '耗时：', time.time() - start_time)

    start_time = time.time()
    score_ = compare_dhash(img1, img2, resize=(200, 200))
    print('感知哈希的相似度：', score_, '耗时：', time.time() - start_time)

    start_time = time.time()
    score_ = compare_hist(img1, img2, resize=(200, 200))
    print('直方图的相似度：', score_, '耗时：', time.time() - start_time)






# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Function:
import numpy as np
import scipy.spatial.distance as dist
from sklearn.metrics.pairwise import cosine_similarity


class Cosine:

    def __init__(self):
        self.func = self.cosine_vector_optimization
        pass

    def run(self, arr1: np.array, arr2: np.array):
        return self.func(arr1, arr2)

    @staticmethod
    def cosine_vector_base(arr1: np.array, arr2: np.array):
        """
        Efficiency is lowest: 122500 times needs 38.62 second.
        """
        return cosine_similarity(X=[arr1], Y=[arr2])[0][0]

    @staticmethod
    def cosine_vector_with_matrix(arr: np.array, filter_result=False, filter_threshold: float = 0):
        """
        Efficiency is high: 122500 times needs 1.3404 second.
        Input data:
            X = np.array([vector1, vector2, vector3])
            X : {array-like, sparse matrix} of shape (n_samples_X, n_features)
        filter_result:
        filter_threshold:
        Return a matrix:
            similarities[0, 1]: item1 & item2
            similarities[0, 2]: item1 & item3
        OR return a list:
            [[index1, index2, value], [index1, index3, value]]
        """
        similarity_matrix = cosine_similarity(arr)
        if filter_threshold > 0:
            similarity_matrix = np.where(similarity_matrix < filter_threshold, 0, similarity_matrix)
        if filter_result:
            res = []
            upper_triangular = np.triu(similarity_matrix, k=1)
            nonzero = np.nonzero(upper_triangular)
            for i in range(len(nonzero[0])):
                res.append((nonzero[0][i], nonzero[1][i], upper_triangular[nonzero[0][i]][nonzero[1][i]]))
            return res
        return similarity_matrix

    @staticmethod
    def cosine_vector_optimization(arr1: np.array, arr2: np.array):
        """
        Parallel optimization based on CPU
        Efficiency is highest: 122500 times needs 0.6753 second.
        """
        return 1 - dist.cosine(arr1, arr2)




# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Function: 相似度计算
import re
import jieba
from typing import List
from zhousflib.util import re_util
from zhousflib.metrics.cosine import Cosine
from zhousflib.ml.feature_vector import FeatureVector, TypeFeatureVector

"""
字符串匹配算法：这是最基本的文本相似度计算方法，主要通过将两个文本字符串进行逐个字符的比较，计算相同字符的数量占总字符数的比例来判断文本的相似度。但是，这种方法对于大量文本的比对速度较慢，且只能检测出完全相同的文本
哈希算法：哈希算法可以快速计算出文本的哈希值，然后通过比对哈希值来判断文本的相似度。但是，哈希算法存在哈希冲突的问题，即不同的文本可能会产生相同的哈希值，从而导致误判
N-gram算法：N-gram算法是一种基于文本分词的方法，将文本分成N个连续的词组，然后比对词组的相似度来判断文本的相似度。N-gram算法可以识别出部分相似的文本，相对于字符串匹配算法和哈希算法，其检测精度更高。
向量空间模型算法：向量空间模型算法是一种基于文本向量化的方法，将文本转换成向量，然后计算向量之间的相似度来判断文本的相似度。这种方法可以识别出语义相似的文本，相对于其他算法，其检测精度更高。

MinHash算法：MinHash算法是一种基于哈希的文本查重方法，它通过随机排列文档中的词项并使用哈希函数来比较文档的相似性。
SimHash算法：是一种局部敏感哈希（Locality Sensitive Hashing, LSH）技术，最初由Google提出，用于高效地计算文本相似度
。其核心思想是将相似的文本映射到相近的哈希空间中，从而实现快速的相似性检测。

运行速度：KSentence > Simhash > Minhash
准确率：KSentence > Minhash > Simhash
召回率：Simhash > Minhash > KSentence
工程应用上，海量文本用Simhash，短文本用Minhash，追求速度用KSentence。

余弦相似度：from sklearn.metrics.pairwise import cosine_similarity   
欧氏距离：  from sklearn.metrics.pairwise import euclidean_distances
曼哈顿距离：from sklearn.metrics.pairwise import manhattan_distances
"""


def text_to_vector(text: List[str], vector_type=TypeFeatureVector.TYPE_COUNT_VECTOR):
    return FeatureVector(vector_type=vector_type).fit_transform(text)


def compute_similarity(text: List[str], vector_type=TypeFeatureVector.TYPE_COUNT_VECTOR, filter_punctuation=True, filter_s=True,
                       cut_all=True, filter_result=False, filter_threshold: float = 0):
    """

    :param text:
    :param vector_type:
    :param filter_punctuation: 过滤标点符号
    :param filter_s: 中文计算时，过滤空格、制表符、换行符，若计算英文时filter_s设置为False
    :param cut_all:
    :param filter_result:
    :param filter_threshold:
    :return:
    """
    _text = []
    for txt in text:
        if filter_punctuation:
            txt = re_util.remove_special_punctuation(str(txt))
            if filter_s:
                # 提取中文、数字、字母
                txt = re.sub(r"[^\u0041-\u005a\u0061-\u007a\u0030-\u0039\u4e00-\u9fa5]", "", txt)
            else:
                # 提取中文、数字、字母、包括 \s （空格、制表符、换行符等）
                txt = re.sub(r"[^\u0041-\u005a\u0061-\u007a\u0030-\u0039\u4e00-\u9fa5\s]", "", txt)
        _text.append(str(jieba.lcut(txt, cut_all=cut_all)))
    cosine = Cosine()
    vector = text_to_vector(_text, vector_type)
    similarity_matrix = cosine.cosine_vector_with_matrix(vector, filter_result=filter_result, filter_threshold=filter_threshold)
    return similarity_matrix


if __name__ == "__main__":
    import time
    start = time.time()
    documents = ["This is the first document",
                 "This document is the second document",
                 "This is the third document"]
    documents = ["本项目位于河北省邯郸市河北峰煤焦化公司厂区内本阶段暂利用气象专业软件获取项目所在地太阳辐射数据根据查取专业软件项目场址代表年总辐射量为根据太阳能资源评估方法中太阳能资源的分级评估方法该区域的太阳能资源丰富程度属类区即资源丰富年辐射稳定程度为类稳定年平均直射比散射辐射较多能保证项目有较高的发电量和较好的开发前景"
                 ,"根据查取专业软件项目场址代表年总辐射量为根据太阳能资源评估方法中太阳能资源的分级评估方法该区域的太阳能资源丰富程度属类区即资源丰富年辐射稳定程度为类稳定年平均直射比散射辐射较多能保证项目有较高的发电量和较好的开发前景"]
    results = compute_similarity(text=documents, vector_type=TypeFeatureVector.TYPE_COUNT_VECTOR,
                                 filter_punctuation=True, filter_s=False, cut_all=False, filter_result=True, filter_threshold=0.01)
    print(results)
    print("耗时", time.time() - start)




# -*- coding: utf-8 -*-
# @Author  : zhousf-a
# @Function:
import colorsys
from typing import Any, List, Tuple


def generate_rgb_colors(count: int) -> List:
    """
    生成RGB颜色列表
    :param count: 生成颜色数量
    :return:  [(255, 0, 0), (203, 255, 0), (0, 255, 102), (0, 102, 255), (204, 0, 255)]
    """
    hsv_tuples = [(1.0 * x / count, 1., 1.) for x in range(count)]
    colors = list(map(lambda x: colorsys.hsv_to_rgb(*x), hsv_tuples))
    colors = list(map(lambda x: (int(x[0] * 255), int(x[1] * 255), int(x[2] * 255)), colors))
    return colors


def rgb_to_hex(rgb: Any):
    if isinstance(rgb, Tuple):
        return '#{:02x}{:02x}{:02x}'.format(rgb[0], rgb[1], rgb[2])
    if isinstance(rgb, List):
        return ['#{:02x}{:02x}{:02x}'.format(color[0], color[1], color[2]) for color in rgb]


def generate_hex_colors(count: int):
    """
    生成HEX颜色列表
    :param count: 生成颜色数量
    :return: ['#ff0000', '#cbff00', '#00ff66', '#0066ff', '#cc00ff']
    """
    rgb_colors = generate_rgb_colors(count)
    return rgb_to_hex(rgb_colors)






# -*- coding:utf-8 -*-
# Author:  zhousf
# Description:
import cv2
import numpy as np
from pathlib import Path

"""
图像的高频信息、低频信息
低频信息：代表着图像中亮度/灰度值/颜色变化很缓慢的区域，描述了图像的主要部分，是对整幅图像强度的综合度量
高频信息：对应着图像变化剧烈的部分，也就是图像的边缘/轮廓、噪声以及细节部分，主要是对图像边缘/轮廓的度量，而人眼对高频分量比较敏感
"""


def read(img_path: Path, bg_to_white=False):
    """
    读图片-兼容图片路径包含中文
    :param img_path:
    :param bg_to_white: 是否将图片的透明背景转成白色背景
    :return: np.ndarray
    """
    if isinstance(img_path, str):
        img_path = Path(img_path)
    if isinstance(img_path, Path):
        if bg_to_white:
            img_path = cv2.imdecode(np.fromfile(img_path, dtype=np.uint8), cv2.IMREAD_UNCHANGED)
            if img_path is not None and img_path.shape[2] == 4:
                alpha_channel = img_path[:, :, 3]
                if cv2.countNonZero(alpha_channel) < alpha_channel.size:
                    b_channel, g_channel, r_channel, a_channel = cv2.split(img_path)
                    white_background = np.ones_like(a_channel) * 255
                    a_channel = a_channel / 255.0
                    r_channel = r_channel * a_channel + white_background * (1 - a_channel)
                    g_channel = g_channel * a_channel + white_background * (1 - a_channel)
                    b_channel = b_channel * a_channel + white_background * (1 - a_channel)
                    img_path = cv2.merge((b_channel, g_channel, r_channel))
        else:
            img_path = cv2.imdecode(np.fromfile(str(img_path), dtype=np.uint8), cv2.IMREAD_COLOR)
    return img_path


def write(image: np.ndarray, img_write_path: Path):
    """
    写图片-兼容图片路径包含中文
    :param image:
    :param img_write_path:
    :return:
    """
    cv2.imencode(img_write_path.suffix, image[:, :, ::-1])[1].tofile(str(img_write_path))


def is_transparent(img_path: Path):
    """
    判断图片是否透明背景
    :param img_path:
    :return:
    """
    img = cv2.imdecode(np.fromfile(str(img_path), dtype=np.uint8), cv2.IMREAD_UNCHANGED)
    if img is not None and img.shape[2] == 4:
        alpha_channel = img[:, :, 3]
        if cv2.countNonZero(alpha_channel) < alpha_channel.size:
            return True
    return False


def transparent_bg_to_white(img_path: Path, save_path: Path = None):
    """
    图片透明背景转成白色背景
    :param img_path:
    :param save_path:
    :return:
    """
    img = cv2.imdecode(np.fromfile(img_path, dtype=np.uint8), cv2.IMREAD_UNCHANGED)
    if img is not None and img.shape[2] == 4:
        alpha_channel = img[:, :, 3]
        if cv2.countNonZero(alpha_channel) < alpha_channel.size:
            b_channel, g_channel, r_channel, a_channel = cv2.split(img)
            white_background = np.ones_like(a_channel) * 255
            a_channel = a_channel / 255.0
            r_channel = r_channel * a_channel + white_background * (1 - a_channel)
            g_channel = g_channel * a_channel + white_background * (1 - a_channel)
            b_channel = b_channel * a_channel + white_background * (1 - a_channel)
            result = cv2.merge((b_channel, g_channel, r_channel))
            if save_path is not None:
                write(result, save_path)
            return result
    return img




# -*- coding:utf-8 -*-
# Author:      zhousf
# Description:  批量异步下载工具类
"""
pip install grequests
"""
from pathlib import Path

import grequests
from requests import Response


class DownloadBatch(object):

    def __init__(self, concurrent=True):
        """
        批量下载文件工具类
        :param concurrent: True并行  False串行
        """
        self.req_list = []
        self.consume_time = 0
        self.file_names = []
        self.task_num = 0
        self.concurrent = concurrent

    @staticmethod
    def exception_handler(request, exception):
        r = Response()
        r.status_code = 408
        r.reason = "download failed: {0}".format(exception)
        return r

    def add(self, file_name, url, timeout=20.0):
        if url is None or file_name is None:
            return False
        if file_name in self.file_names:
            return False
        self.file_names.append(file_name)
        self.req_list.append(grequests.get(url, timeout=timeout))
        return True

    def add_all(self, files):
        """
        :param files: [{"name": "1.jpg", "url": ""}]
        :return:
        """
        for file in files:
            name = file.get("name")
            url = file.get("url")
            if url is None or name is None:
                continue
            timeout = file.get("timeout", default=20.0)
            self.req_list.append(grequests.get(url, timeout=timeout))
            self.file_names.append(name)

    def run(self, save_dir: Path):
        download_result = []
        try:
            if not save_dir.exists():
                save_dir.mkdir(parents=True)
            self.task_num = len(self.req_list)
            if self.task_num == 0:
                return download_result
            size = self.task_num if self.concurrent else 1
            responses = grequests.map(requests=self.req_list, size=size, exception_handler=self.exception_handler)
            for i in range(0, len(responses)):
                response = responses[i]
                try:
                    save_file = str(save_dir.joinpath(self.file_names[i]))
                    # target_type = self.file_names[i].split(".")[-1]
                    # if 'Content-Type' in response.headers:
                    #     content_type = response.headers['Content-Type']
                    #     if target_type in ["jpg", "JPG", "JPEG", "jpeg", "png", "PNG", "gif", "GIF"]:
                    #         target_type = "image"
                    #     if content_type.find(target_type) < 0:
                    #         download_result.append((False,
                    #                                 "The file type is {0}, but {1} is expected".format(
                    #                                                    content_type,
                    #                                                    target_type),
                    #                                 self.file_names[i],
                    #                                 self.req_list[i].url))
                    #         continue
                    if response.status_code != 200:
                        download_result.append((
                            False,
                            "{0} {1} {2}".format(self.file_names[i], response.reason, response.status_code),
                            self.file_names[i],
                            self.req_list[i].url
                        ))
                        continue
                    # 当信息流小于100字节，则不是文件
                    if len(response.text) <= 100:
                        download_result.append((
                            False,
                            "{0} {1} {2}".format(self.file_names[i], response.reason, response.status_code),
                            self.file_names[i],
                            self.req_list[i].url
                        ))
                        continue
                    with open(save_file, "wb") as f:
                        f.write(response.content)
                        download_result.append((
                            True,
                            save_file,
                            self.file_names[i],
                            self.req_list[i].url
                        ))
                except Exception as ex:
                    download_result.append((
                        False,
                        "{0} {1} {2}".format(self.file_names[i], response.reason, response.status_code),
                        self.file_names[i],
                        self.req_list[i].url
                    ))
                    continue
        finally:
            self.req_list.clear()
            self.task_num = 0
            self.file_names.clear()
        return download_result

    def run_one_file(self, save_dir: Path, file_name, url, timeout=20.0):
        self.add(file_name=file_name, url=url, timeout=timeout)
        results = self.run(save_dir=save_dir)
        return results[0]


if __name__ == "__main__":
    downloader = DownloadBatch(concurrent=True)
    res = downloader.run_one_file(save_dir=Path(r"C:\Users\zhousf-a\Desktop\download"),
                                  file_name="input.zip",
                                  url="")



# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Function: 不支持多进程
"""
pip install tinydb
"""
from pathlib import Path
from tinydb import TinyDB, Query
from tinydb.table import Document
from typing import (
    List,
    Mapping,
    TypeVar
)


class Entity:

    def __init__(self, document: Document = None):
        self.doc_id = None
        if document is not None:
            self.__dict__ = document

    def __call__(self, *args, **kwargs):
        pass

    def update(self, document: Document = None):
        if document is not None:
            self.__dict__.update(document)
            self.doc_id = document.doc_id
        return self

    def mapping(self):
        dic = self.__dict__
        if "doc_id" in dic:
            dic.pop("doc_id")
        return dic


T = TypeVar("T", bound=Entity)


class TINYDB:
    def __init__(self, db_dir: Path, db_name="db"):
        self.db = TinyDB(db_dir.joinpath("{0}.json".format(db_name)))

    def insert(self, document: Mapping) -> int:
        return self.db.insert(document)

    def insert_if_not_exist(self, document: Mapping, cond=None, doc_id=None) -> int:
        if self.db.contains(cond, doc_id):
            return -1
        return self.db.insert(document)

    def contains(self, cond=None, doc_id=None) -> bool:
        return self.db.contains(cond, doc_id)

    def search(self, cond=None) -> list:
        return self.db.search(cond)

    def query(self, obj: T, cond=None) -> List[T]:
        if cond:
            docs = self.db.search(cond)
        else:
            docs = self.db.all()
        return [obj.update(i) for i in docs]

    def query_all(self) -> List[Document]:
        return self.db.all()

    def update(self, fields, cond=None, doc_ids=None) -> List[int]:
        return self.db.update(fields, cond, doc_ids)

    def count(self, cond) -> int:
        return self.db.count(cond)

    def remove(self, cond, doc_ids) -> List[int]:
        return self.db.remove(cond, doc_ids)

    def remove_all(self):
        self.db.truncate()

    @staticmethod
    def query_ins():
        return Query()


if __name__ == "__main__":
    # from configure import DBDir
    # db = TINYDB(Path(__file__).parent, "test")

    # data = db.query_all()
    # print(data)

    # db.remove_all()
    # print(db.insert_if_not_exist({'task_id': 3, 'task_dir': '/root/image', 'task_status': 2}, db.query_ins().task_id == 3))

    # for item in db.query(obj=Entity()):
    #     print(item.__dict__)

    # result_ = db.query_all_return_class(obj=Note())
    # print(result_[0].task_id)

    # print(db.update(fields={"task_status": 0}, cond=None, doc_ids=[result.doc_id]))
    # print(db.update(fields={"task_status": 2}, cond=None, doc_ids=[result.doc_id]))
    pass




# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Function:
import sys
import copy
from pathlib import Path
from typing import Optional

import cv2
import yaml
import numpy
from PIL import Image
import fastdeploy as fd
from easydict import EasyDict
from fastdeploy import ModelFormat
from fastdeploy.libs.fastdeploy_main.vision import (
    ClassifyResult,
    DetectionResult,
    SegmentationResult,
    OCRResult
)

from zhousflib.image import read
from zhousflib.image import op
from zhousflib.image import pil_util
from zhousflib.util import ocr_vis_util
from zhousflib.font import Font_SimSun
from zhousflib.image import write
from zhousflib.infer_framework.fast_infer.backend import Backend
"""
download wheel
https://www.paddlepaddle.org.cn/whl/fastdeploy.html

https://www.paddlepaddle.org.cn/install/quick?docurl=/documentation/docs/zh/install/pip/linux-pip.html

version base on cpu
pip install fastdeploy-python -f https://www.paddlepaddle.org.cn/whl/fastdeploy.html

version base on gpu
pip install fastdeploy-gpu-python -f https://www.paddlepaddle.org.cn/whl/fastdeploy.html
conda config --add channels conda-forge && conda install cudatoolkit=11.2 cudnn=8.2

[question]
ImportError: libcudart.so.11.0: cannot open shared object file: No such file or directory
[solution]
cd $CONDA_PREFIX
mkdir -p ./etc/conda/activate.d
touch ./etc/conda/activate.d/env_vars.sh

[question]
ImportError: libGL.so.1: cannot open shared object file: No such file or directory
[solution]
pip install opencv-python-headless -i https://pypi.tuna.tsinghua.edu.cn/simple

[question]
RuntimeError: FastDeploy initalized failed! Error: /lib64/libstdc++.so.6: version `GLIBCXX_3.4.21' not found
[solution]
find / -name "libstdc++.so*"
找到后选择其中一个目录， 例如：/root/anaconda3/lib
export LD_LIBRARY_PATH=/root/anaconda3/lib/:$LD_LIBRARY_PATH


"""


class BackendFastDeploy(Backend):
    def __init__(self, *args, **kwargs):
        self.model_dir: Path = kwargs.get("model_dir")
        self.device_id = kwargs.get("device_id")
        self.backend = None
        self.plugin = None
        self.model = None
        self.label_list = []
        self.clone_model = False
        super().__init__(*args)
        self.plugins = self.__register_plugin()

    @staticmethod
    def __register_plugin():
        plugin_classification = {
            "fd.vision.classification.PaddleClasModel": fd.vision.classification.PaddleClasModel,
            "fd.vision.classification.PPShiTuV2Recognizer": fd.vision.classification.PPShiTuV2Recognizer,
        }
        plugin_detection = {
            "fd.vision.detection.PaddleDetectionModel": fd.vision.detection.PaddleDetectionModel,
            "fd.vision.detection.PPYOLOE": fd.vision.detection.PPYOLOE,
            "fd.vision.detection.PPYOLO": fd.vision.detection.PPYOLO,
            "fd.vision.detection.PPYOLOER": fd.vision.detection.PPYOLOER,
            "fd.vision.detection.PaddleYOLOX": fd.vision.detection.PaddleYOLOX,
            "fd.vision.detection.YOLOX": fd.vision.detection.YOLOX,
            "fd.vision.detection.YOLOv8": fd.vision.detection.YOLOv8,
            "fd.vision.detection.YOLOv7": fd.vision.detection.YOLOv7,
            "fd.vision.detection.YOLOv6": fd.vision.detection.YOLOv6,
        }
        plugin_segmentation = {
            "fd.vision.segmentation.PaddleSegModel": fd.vision.segmentation.PaddleSegModel,
        }
        plugin_ocr = {
            "fd.vision.ocr.DBDetector": fd.vision.ocr.DBDetector,
            "fd.vision.ocr.Classifier": fd.vision.ocr.Classifier,
            "fd.vision.ocr.Recognizer": fd.vision.ocr.Recognizer,
            "fd.vision.ocr.PPOCRv4": fd.vision.ocr.PPOCRv4,
            "fd.vision.ocr.PPOCRv3": fd.vision.ocr.PPOCRv3,
        }
        plugin_text = {
            "fd.text.uie.UIEModel": fd.text.uie.UIEModel
        }
        if sys.version_info > (3, 9):
            return plugin_classification | plugin_detection | plugin_segmentation | plugin_ocr | plugin_text
        elif (3, 5) < sys.version_info < (3, 9):
            return {**plugin_classification, **plugin_detection, **plugin_segmentation, **plugin_ocr, **plugin_text}
        else:
            merged_dict = dict(plugin_classification, **plugin_detection)
            merged_dict = dict(merged_dict, **plugin_segmentation)
            merged_dict = dict(merged_dict, **plugin_ocr)
            merged_dict = dict(merged_dict, **plugin_text)
            return merged_dict

    def build(self, **kwargs):
        plugin_name = kwargs.get("plugin")
        clone_model = kwargs.get("clone_model", False)
        self.plugin = self.plugins.get(plugin_name, None)
        if self.plugin is None:
            raise Exception("Plugin not found: {0}".format(self.plugin))
        if "plugin" in kwargs:
            kwargs.pop("plugin")
        if "clone_model" in kwargs:
            kwargs.pop("clone_model")
        runtime_option = kwargs.get("runtime_option", None)
        if runtime_option is None:
            runtime_option = fd.RuntimeOption()
            runtime_option.use_paddle_backend()
        if self.device_id >= 0:
            runtime_option.use_gpu(device_id=self.device_id)
        else:
            runtime_option.use_cpu()
        model_format = kwargs.get("model_format", ModelFormat.PADDLE)
        if self.model_dir is not None:
            txt_file_mapping = {
                "fd.vision.ocr.Recognizer": "label_path",
                "fd.text.uie.UIEModel": "vocab_file",
            }
            union = {
                "model_file": self.get_file_path_by_suffix(model_dir=self.model_dir, suffix=".pdmodel"),
                "params_file": self.get_file_path_by_suffix(model_dir=self.model_dir, suffix=".pdiparams"),
                "config_file": self.get_file_path_by_suffix(model_dir=self.model_dir, suffix=".yaml"),
                txt_file_mapping.get(plugin_name): self.get_file_path_by_suffix(model_dir=self.model_dir, suffix=".txt"),
                "runtime_option": runtime_option,
                "model_format": model_format
            }
            union = {k: v for k, v in union.items() if v is not None}
            if sys.version_info > (3, 9):
                union = union | kwargs
            elif (3, 5) < sys.version_info < (3, 9):
                union = {**union, **kwargs}
            else:
                union = dict(union, **kwargs)
            try:
                self.model = self.plugin(**union)
            except Exception as e:
                raise Exception("Please check model_dir files: {0} {1}".format(self.model_dir, e))
        else:
            if self.plugin in [fd.vision.ocr.PPOCRv4, fd.vision.ocr.PPOCRv3]:
                det_model = kwargs.get("det_model", None)
                cls_model = kwargs.get("cls_model", None)
                rec_model = kwargs.get("rec_model", None)
                self.model = self.plugin(det_model, cls_model, rec_model)

        if clone_model:
            if hasattr(self.model, "runtime_option"):
                opt = self.model.runtime_option
                if hasattr(opt, "backend"):
                    backend_ = opt.backend
                    if hasattr(backend_, "name"):
                        if backend_.name != "ORT":
                            self.clone_model = True

        if self.model_dir is not None:
            # init label list
            config_file = self.get_file_path_by_suffix(model_dir=self.model_dir, suffix=".yaml")
            if config_file is not None and isinstance(config_file, str):
                with Path(config_file).open("r") as f:
                    yml_conf = yaml.safe_load(f)
                    if "label_list" in yml_conf:
                        self.label_list = yml_conf["label_list"]
            if len(self.label_list) == 0:
                label_file = self.get_file_path_by_suffix(model_dir=self.model_dir, suffix=".label")
                if label_file is not None and isinstance(label_file, str):
                    with Path(label_file).open("r", encoding="utf-8") as f:
                        for label in f.readlines():
                            label = label.replace("\n", "")
                            item = label.split(" ")
                            if len(item) == 2:
                                self.label_list.append(item[1])

    @staticmethod
    def draw_boxes(infer_result, **kwargs):
        if not hasattr(infer_result, "boxes"):
            return
        boxes = infer_result.boxes
        if len(boxes) == 0:
            return
        arr = numpy.asarray(boxes)
        if len(arr.shape) == 2 and arr.shape[1] == 8:
            arr = arr.reshape([arr.shape[0], 4, 2])
            boxes = arr.tolist()
        # config of visualization
        vis_font = kwargs.get("vis_font", None)
        vis_show = kwargs.get("vis_show", False)
        vis_image_size = kwargs.get("vis_image_size", None)
        vis_fill_transparent = kwargs.get("vis_fill_transparent", 60)
        # config of save visualization file
        vis_image_file: Optional[Path] = kwargs.get("vis_image_file", None)

        image_file = kwargs.get("image_path", None)
        if vis_image_file is not None or vis_show:
            draw_img = pil_util.draw_polygon(polygon=boxes,
                                             image_file=image_file,
                                             image_size=vis_image_size,
                                             font=vis_font,
                                             fill_transparent=vis_fill_transparent,
                                             show=vis_show)
            if vis_image_file is not None:
                draw_img.save(vis_image_file)

    def op_classify(self, infer_result, image_arr, **kwargs):
        result = EasyDict()
        label_names = []
        label_ids = infer_result.label_ids
        scores = infer_result.scores
        for i, score in enumerate(scores):
            label_name = self.label_list[label_ids[i]]
            label_names.append(label_name)
        result.feature = infer_result.feature
        result.label_ids = infer_result.label_ids
        result.label_names = label_names
        result.scores = infer_result.scores
        return result

    def op_detection(self, infer_result, image_arr, **kwargs):
        result = EasyDict()
        label_ids = infer_result.label_ids
        scores = infer_result.scores
        bboxes = []
        items = []
        label_names = []
        for i, box in enumerate(infer_result.boxes):
            bboxes.append([int(box[0]), int(box[1]), int(box[2]), int(box[3])])
            class_name = self.label_list[label_ids[i]]
            items.append([class_name, scores[i], int(box[0]), int(box[1]), int(box[2]), int(box[3])])
            label_names.append(class_name)
        result.items = items
        result.label_ids = infer_result.label_ids
        result.label_names = label_names
        result.boxes = bboxes
        result.scores = infer_result.scores
        result.masks = infer_result.masks
        result.contain_masks = infer_result.contain_masks
        self.draw_boxes(infer_result, **kwargs)
        return result

    @staticmethod
    def op_segmentation(infer_result, image_arr, **kwargs):
        result = EasyDict()
        result.label_map = infer_result.label_map
        result.contain_score_map = infer_result.contain_score_map
        result.score_map = infer_result.score_map
        result.shape = infer_result.shape
        if len(result.label_map) > 0:
            # 是否开启掩码最大连通域计算
            max_connectivity_domain = kwargs.get("max_connectivity_domain", False)
            # 掩码空洞填充阈值，面积占整图比例，0则不填充，小于该阈值时则填充
            mask_hole_area_thresh = kwargs.get("mask_hole_area_thresh", 0)
            arr_mask = numpy.array(result.label_map).reshape(result.shape[0], result.shape[1])
            """
            最大连通域计算
            """
            if max_connectivity_domain:
                max_region = op.max_connectivity_domain(mask_arr=arr_mask)
                result.label_map = max_region.flatten().tolist()
            """
            去掉连通域中的空洞
            """
            if mask_hole_area_thresh > 0:
                image_area = image_arr.shape[0] * image_arr.shape[1]
                arr_mask = numpy.array(result.label_map).reshape(result.shape[0], result.shape[1])
                img_max = numpy.where(arr_mask == 1, 0, 1) * 255
                # 掩码类型转换
                img_max = img_max.astype(dtype=numpy.uint8)
                contours, _ = cv2.findContours(img_max, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
                cv_contours = []
                for contour in contours:
                    area = cv2.contourArea(contour)
                    if area / image_area <= mask_hole_area_thresh:
                        cv_contours.append(contour)
                    else:
                        continue
                if len(cv_contours):
                    cv2.fillPoly(img_max, cv_contours, (0, 0, 0))
                    img_max = numpy.where(img_max == 0, 1, 0)
                    img_max = img_max.astype(dtype=numpy.int32)
                    result.label_map = img_max.flatten().tolist()
            vis_show = kwargs.get("vis_show", False)
            vis_transparent_weight = kwargs.get("vis_transparent_weight", 0.6)
            vis_image_file: Optional[Path] = kwargs.get("vis_image_file", None)
            if vis_image_file is not None or vis_show:
                im_vis = fd.vision.visualize.vis_segmentation(image_arr, infer_result, vis_transparent_weight)
                pil_image = Image.fromarray(im_vis)
                if vis_image_file is not None:
                    pil_image.save(vis_image_file, quality=100)
                if vis_show:
                    pil_image.show()
        return result

    def op_ocr(self, infer_result, image_arr, **kwargs):
        result = EasyDict()
        result.boxes = infer_result.boxes
        result.cls_labels = infer_result.cls_labels
        result.cls_scores = infer_result.cls_scores
        result.rec_scores = infer_result.rec_scores
        result.text = infer_result.text
        if len(result.text) == 0 and len(result.boxes) > 0:
            self.draw_boxes(infer_result, **kwargs)
        if len(result.text) > 0 and len(result.boxes) > 0:
            arr = numpy.asarray(result.boxes)
            if len(arr.shape) == 2 and arr.shape[1] == 8:
                arr = arr.reshape([arr.shape[0], 4, 2])
                result.boxes = arr.tolist()
            vis_image_file: Optional[Path] = kwargs.get("vis_image_file", None)
            vis_show = kwargs.get("vis_show", False)
            if vis_image_file is not None or vis_show:
                boxes = []
                texts = []
                scores = []
                for i, bbox in enumerate(result.boxes):
                    if len(bbox) == 0:
                        continue
                    bbox_arr = numpy.asarray(bbox)
                    boxes.append(bbox_arr)
                    texts.append(result.text[i])
                    scores.append(result.rec_scores[i])
                image_file: Optional[Path] = kwargs.get("image_path", None)
                image = Image.open(image_file)
                draw_img = ocr_vis_util.draw_ocr_box_txt(
                    image,
                    boxes,
                    texts,
                    scores,
                    drop_score=0.5,
                    font_path=str(Font_SimSun))
                if vis_image_file is not None:
                    write(image=draw_img, img_write_path=vis_image_file)
                if vis_show:
                    draw_img = Image.fromarray(draw_img)
                    draw_img.show()
        return result

    def post_process(self, infer_result, image_arr, **kwargs):
        operations = {
            ClassifyResult: lambda: self.op_classify(infer_result, image_arr, **kwargs),
            DetectionResult: lambda: self.op_detection(infer_result, image_arr, **kwargs),
            SegmentationResult: lambda: self.op_segmentation(infer_result, image_arr, **kwargs),
            OCRResult: lambda: self.op_ocr(infer_result, image_arr, **kwargs),
        }
        return operations.get(type(infer_result), lambda: infer_result)()

    def inference(self, input_data, **kwargs):
        image_arr = read(input_data)
        model = self.model.clone() if self.clone_model and hasattr(self.model, "clone") else self.model
        schema = kwargs.get("schema", None)
        kwargs_infer = copy.deepcopy(kwargs)
        delete_key = [k for k in kwargs_infer.keys() if k in ["vis_image_file", "vis_show", "vis_font", "vis_image_size",
                                                              "vis_fill_transparent", "vis_transparent_weight", "schema"]]
        if len(delete_key) > 0:
            for k in delete_key:
                kwargs_infer.pop(k)
        if hasattr(model, "set_schema") and schema:
            self.model.set_schema(schema)
        infer_res = model.predict(image_arr, **kwargs_infer)

        if isinstance(input_data, Path):
            kwargs["image_path"] = input_data
        res = self.post_process(infer_res, image_arr, **kwargs)
        return res
